import yaml

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader


from django.apps import apps
from django.db import models
from django.core.exceptions import FieldDoesNotExist
from django.core.serializers.base import DeserializationError
from django.core.serializers.python import Serializer as PythonSerializer
from django.core.serializers.pyyaml import DjangoSafeDumper

from .models import parse_dump_path
from .python_serializer import Deserializer as PythonDeserializer


class Serializer(PythonSerializer):
    internal_use_only = False

    def end_serialization(self):
        assert len(self.objects) == 1
        obj = self.objects[0]

        app_label, model_name = obj['model'].split('.')
        model = apps.get_model(app_label, model_name)

        fields = {name: value for name, value in obj['fields'].items() if value}

        fields.pop('key', None)
        fields.pop('content_format', None)
        content = fields.pop('content', None)

        if model.key_structure is not None:
            for field_name in model.field_names_from_key_structure():
                fields.pop(field_name)

        for field_name, field_value in fields.items():
            if is_fk_field(model, field_name):
                assert isinstance(field_value, tuple) and len(field_value) == 1
                fields[field_name] = field_value[0]

            if is_m2m_field(model, field_name):
                assert isinstance(field_value, list) and all(len(v) == 1 for v in field_value)
                fields[field_name] = [v[0] for v in field_value]

            if is_time_field(model, field_name):
                # See comment in django.core.serializers.pyyaml.Serializer.handle_field
                fields[field_name] = str(fields[field_name])

        yaml.dump(fields, self.stream, Dumper=DjangoSafeDumper,
                  default_flow_style=False, **self.options)

        if content is not None:
            self.stream.write('---\n')
            self.stream.write(content)

    def getvalue(self):
        return super(PythonSerializer, self).getvalue()


# Built-in deserializers take either a stream or string, but we require a file,
# because we extract some of the data about the object deserialized in the file
# from its path.
def Deserializer(file, **options):
    model, key, content_format = parse_dump_path(file.name)

    fields = {'key': key}
    fields.update(model.fields_from_key(key))

    data = file.read().decode('utf-8')
    separator = '\n---\n'
    parts = data.split(separator, 1)

    if model.has_content:
        try:
            yaml_fields = yaml.load(parts[0], Loader=SafeLoader)
        except yaml.YAMLError as e:
            raise DeserializationError(e)

        if len(parts) == 1:
            if isinstance(yaml_fields, dict):
                raise DeserializationError('Missing content')

            fields['content'] = parts[0]
        else:
            try:
                fields.update(yaml.load(parts[0], Loader=SafeLoader))
            except yaml.YAMLError as e:
                raise DeserializationError(e)
            fields['content'] = parts[1]

        fields['content_format'] = content_format
    else:
        assert len(parts) == 1
        try:
            fields.update(yaml.load(parts[0], Loader=SafeLoader))
        except yaml.YAMLError as e:
            raise DeserializationError(e)

    for field_name, field_value in fields.items():
        if is_fk_field(model, field_name):
            assert isinstance(field_value, str)
            fields[field_name] = [field_value]

        if is_m2m_field(model, field_name):
            assert isinstance(field_value, list) and all(isinstance(v, str) for v in field_value)
            fields[field_name] = [[v] for v in field_value]

        if is_time_field(model, field_name):
            if isinstance(fields[field_name], int):
                num_seconds = fields[field_name]
                assert 0 <= num_seconds <= 24 * 60 * 60
                hours, minutes_and_seconds = divmod(num_seconds, 60 * 60)
                minutes, seconds = divmod(minutes_and_seconds, 60)
                fields[field_name] = '{}:{}:{}'.format(hours, minutes, seconds)

    record = {
        'model': '{}.{}'.format(model._meta.app_label, model._meta.model_name),
        'fields': fields,
    }

    try:
        yield from PythonDeserializer([record], **options)
    except Exception as e:
        raise DeserializationError(e)


def is_fk_field(model, field_name):
    try:
        field = model._meta.get_field(field_name)
    except FieldDoesNotExist:
        return False

    return field.remote_field and isinstance(field.remote_field, models.ManyToOneRel)


def is_m2m_field(model, field_name):
    try:
        field = model._meta.get_field(field_name)
    except FieldDoesNotExist:
        return False

    return field.remote_field and isinstance(field.remote_field, models.ManyToManyRel)


def is_time_field(model, field_name):
    try:
        field = model._meta.get_field(field_name)
    except FieldDoesNotExist:
        return False

    return isinstance(field, models.TimeField)



import os
from time import sleep

from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.core.management.commands.runserver import Command as RunserverCommand

from ...serialization_helpers import find_file_paths_in_dir, load_from_file
from ...models import DjangoPagesModel, parse_dump_path
from ...utils import run_runserver_in_process


def load_changed(changed_paths):
    load_from_file(changed_paths)

    # TODO  Decide what to do if loading data raises DeserializationError.


def remove_missing(missing_paths):
    for path in missing_paths:
        model, key, _ = parse_dump_path(path)

        instance = model.objects.get_by_natural_key(key)
        instance.delete()

    # TODO  Decide what to do if deleting the object causes cascading
    # deletes.  Options:
    #  * do nothing, leaving data on filesystem in inconsistent state;
    #  * roll back and raise exception;
    #  * remove related files from filesystem.


def get_mtimes():
    mtimes = {}

    for model in DjangoPagesModel.subclasses():
        for path in find_file_paths_in_dir(model.get_dump_dir_path()):
            try:
                stat = os.stat(path)
                mtimes[path] = stat.st_mtime
            except FileNotFoundError:
                # Race condition: the file has disappeared in the time between
                # the glob and the stat.
                pass

    return mtimes


def compare_mtimes(old_mtimes, new_mtimes):
    changed_paths = []
    missing_paths = []

    for path, old_mtime in old_mtimes.items():
        new_mtime = new_mtimes.get(path)
        if new_mtime is None:
            missing_paths.append(path)
        elif new_mtime != old_mtime:
            changed_paths.append(path)

    for path in new_mtimes:
        if path not in old_mtimes:
            changed_paths.append(path)

    return changed_paths, missing_paths


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            'port',
            nargs='?',
            default=RunserverCommand.default_port,
            help='Optional port number'
        )

    def handle(self, *args, **kwargs):
        port = kwargs.get('port')

        call_command('loadpages')

        p = run_runserver_in_process(port)

        try:
            self.serve()
        finally:
            p.terminate()

        print()

    def serve(self):
        mtimes = get_mtimes()

        while True:
            try:
                sleep(0.1)
            except KeyboardInterrupt:
                break

            new_mtimes = get_mtimes()
            changed_paths, missing_paths = compare_mtimes(mtimes, new_mtimes)
            load_changed(changed_paths)
            remove_missing(missing_paths)
            mtimes = new_mtimes



import os
import shutil

import http_crawler

from django.conf import settings
from django.core.management import call_command
from django.core.management.base import BaseCommand

from django_amber.utils import get_free_port, run_runserver_in_process


class Command(BaseCommand):
    def handle(self, *args, **kwargs):
        port = get_free_port()

        p = run_runserver_in_process(port)

        try:
            self.buildsite(port)
        finally:
            p.terminate()

    def buildsite(self, port):
        call_command('loadpages')

        output_path = os.path.join(settings.BASE_DIR, 'output')
        shutil.rmtree(output_path, ignore_errors=True)

        for rsp in http_crawler.crawl('http://localhost:{}/'.format(port)):
            rsp.raise_for_status()

            parsed_url = http_crawler.urlparse(rsp.url)

            if parsed_url.netloc != 'localhost:{}'.format(port):
                # This is an external request, which we don't care about
                continue

            path = parsed_url.path
            segments = path.split('/')
            assert segments[0] == ''
            if segments[-1] == '':
                rel_dir_path = os.path.join(*segments[:-1])
                filename = 'index.html'
            elif '.' in segments[-1]:
                rel_dir_path = os.path.join(*segments[:-1])
                filename = segments[-1]
            else:
                rel_dir_path = os.path.join(*segments)
                filename = 'index.html'

            dir_path = os.path.join(output_path, rel_dir_path)
            os.makedirs(dir_path, exist_ok=True)
            with open(os.path.join(dir_path, filename), 'wb') as f:
                f.write(rsp.content)

        cname = getattr(settings, 'DJANGO_AMBER_CNAME', None)

        if cname:
            with open(os.path.join(output_path, 'CNAME'), 'w') as f:
                f.write(cname)



import os
import re

from django.apps import apps
from django.conf import settings
from django.db import models


class PagesManager(models.Manager):
    def get_by_natural_key(self, key):
        return self.get(key=key)


class DjangoPagesModel(models.Model):
    objects = PagesManager()

    class Meta:
        abstract = True

    dump_dir_path = None
    key_structure = None

    @classmethod
    def get_dump_dir_path(cls):
        if cls.dump_dir_path is None:
            app_config = apps.get_app_config(cls._meta.app_label)
            return os.path.join(app_config.path, 'data', cls._meta.model_name)
        else:
            return os.path.join(settings.BASE_DIR, *cls.dump_dir_path.split('/'))

    @classmethod
    def subclasses(cls):
        for app_config in apps.get_app_configs():
            for model in app_config.get_models():
                if issubclass(model, DjangoPagesModel):
                    yield model

    @classmethod
    def fields_from_key(cls, key):
        if cls.key_structure is None:
            return {}

        pattern = cls.key_structure
        for field_name in cls.field_names_from_key_structure():
            pattern = pattern.replace('[{}]'.format(field_name), '(?P<{}>.+)'.format(field_name))

        match = re.match(pattern, key)
        return match.groupdict()

    @classmethod
    def field_names_from_key_structure(cls):
        return re.findall('\[(\w*)\]', cls.key_structure)

    def dump_path(self):
        if not self.key:
            self.set_key()

        return os.path.join(self.get_dump_dir_path(), *self.key.split('/')) + '.' + self.content_format

    def natural_key(self):
        return (self.key,)

    def set_key(self):
        assert self.key_structure is not None

        key = self.key_structure
        for field_name in self.field_names_from_key_structure():
            value = getattr(self, field_name)
            if isinstance(value, DjangoPagesModel):
                value = value.key
            else:
                value = str(value)
            key = key.replace('[{}]'.format(field_name), value)

        self.key = key

    def __str__(self):
        return self.key


class ModelWithoutContent(DjangoPagesModel):
    key = models.CharField(max_length=255)

    content_format = 'yml'

    has_content = False

    class Meta:
        abstract = True


class ModelWithContent(DjangoPagesModel):
    key = models.CharField(max_length=255)
    content = models.TextField()
    content_format = models.CharField(max_length=255)

    has_content = True

    class Meta:
        abstract = True


def parse_dump_path(path):
    for model in DjangoPagesModel.subclasses():
        if os.path.commonpath([path, model.get_dump_dir_path()]) == model.get_dump_dir_path():
            remainder = os.path.relpath(path, model.get_dump_dir_path())
            key, content_format_with_dot = os.path.splitext(remainder)
            content_format = content_format_with_dot[1:]
            return model, key, content_format

    assert False



from multiprocessing import Process
from time import sleep
from socket import socket

import requests

from django.core.management import call_command
from django.core.management.commands.runserver import Command as RunserverCommand


default_port = RunserverCommand.default_port


def run_runserver_in_process(port=default_port):
    p = Process(
        target=call_command,
        args=('runserver', port),
        kwargs={'use_reloader': False},
    )

    p.start()

    wait_for_server(port)

    return p


def wait_for_server(port=default_port):
    get_with_retries('http://localhost:{}/'.format(port))


def get_with_retries(url, num_retries=5):
    for i in range(num_retries):
        try:
            return requests.get(url)
        except requests.exceptions.ConnectionError:
            pass

        sleep(0.1 * 2 ** i)

    requests.get(url)


def get_free_port():
    s = socket()
    s.bind(('', 0))
    port = s.getsockname()[1]
    s.close()
    return str(port)



import shutil

from django.core.management.base import BaseCommand

from ...models import DjangoPagesModel
from ...serialization_helpers import dump_to_file


class Command(BaseCommand):
    def handle(self, *args, **kwargs):
        for model in DjangoPagesModel.subclasses():
            shutil.rmtree(model.get_dump_dir_path(), ignore_errors=True)
            for obj in model.objects.all():
                dump_to_file(obj)



from django.core.management import call_command
from django.core.management.base import BaseCommand, CommandError

from ...models import DjangoPagesModel
from ...serialization_helpers import find_file_paths_in_dir, load_from_file, LoadFromFileError


class Command(BaseCommand):
    def handle(self, *args, **kwargs):
        call_command('migrate')

        paths = []

        for model in DjangoPagesModel.subclasses():
            paths.extend(find_file_paths_in_dir(model.get_dump_dir_path()))

        try:
            load_from_file(paths)
        except LoadFromFileError as e:
            raise CommandError('Hit error ({}: {}) when loading data from {}'.format(type(e.original_exception), e.original_exception, e.path))



from setuptools import find_packages, setup

import os
import re


def read(*parts):
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, *parts)) as f:
        return f.read()

VERSION = re.search(
    "^__version__ = '(.*)'$",
    read('src', 'django_amber', '__init__.py'),
    re.MULTILINE
).group(1)

if __name__ == '__main__':
    setup(
        name='django-amber',
        version=VERSION,
        description='A Django-powered static site generator',
        long_description=read('README.md'),
        packages=find_packages(where='src'),
        package_dir={'': 'src'},
        install_requires=['Django', 'http-crawler', 'requests', 'PyYAML'],
        url='http://github.com/inglesp/django-amber',
        author='Peter Inglesby',
        author_email='peter.inglesby@gmail.com',
        license='License :: OSI Approved :: MIT License',
    )



import os

from django.db import transaction

from .serializer import Deserializer, Serializer


class LoadFromFileError(Exception):
    def __init__(self, original_exception, path):
        self.original_exception = original_exception
        self.path = path


def dump_to_file(instance):
    dump_path = instance.dump_path()
    os.makedirs(os.path.dirname(dump_path), exist_ok=True)

    serializer = Serializer()
    serializer.serialize([instance], use_natural_foreign_keys=True)
    data = serializer.getvalue()

    with open(dump_path, 'w') as f:
        f.write(data)


def load_from_file(paths):
    objs_with_deferred_fields = []

    with transaction.atomic():
        for path in paths:
            try:
                with open(path, 'rb') as f:
                    for obj in Deserializer(f, handle_forward_references=True):
                        obj.save()

                        if obj.deferred_fields:
                            objs_with_deferred_fields.append(obj)
            except Exception as e:
                raise LoadFromFileError(e, path)

        for obj in objs_with_deferred_fields:
            obj.save_deferred_fields()


def find_file_paths_in_dir(path):
    for root, _, file_paths in os.walk(path):
        for file_path in file_paths:
            if file_path[0] != '.':
                yield os.path.join(root, file_path)
